#  $Id: Classifier.pm,v 1.1.1.1 2008/02/17 23:31:10 jonas Exp $  -*-cperl-*-
package Email::Classifier;

use warnings;
use strict;
use Carp qw( croak confess cluck );
use Module::Pluggable search_path => ['Email::Classifier'];
use Email::Simple::Header;
use vars qw( $AUTOLOAD %PLUGIN );

=head1 NAME

Email::Classifier - Analyzes and classifies emails

=head1 VERSION

Version 0.01

=cut

our $VERSION = '0.01';

=head1 SYNOPSIS

  use Email::Classifier;
  use Email::MIME;

  my $eml = Email::MIME->new($email_string);
  my $msg = Email::Classifier->new($eml);
  if( $msg->is_bounce )
  {
     foreach my $report ( $msg->bounce->reports )
     {
        print $report->as_string;
     }
  }


=head1 DESCRIPTION

Based on L<Mail::DeliveryStatus::BounceParser>

Loads modules under L<Email::Classifier> on demand.

Each moudle exports two methods based on it's name. For example:

L<Email::Classifier::Bounce> exports:

  is_bounce()
  bounce()

The C<$name> method returns the object giving futher access in
respective module.

The C<is_$name> method returns true if the module recognized the email
as the type it handles.

=cut

our $CLASS_HEADER = "Email::Simple::Header";

#######################################################################

=head2 new

  Email::Classifier->new( $email )

C<$email> should be an object compatible with L<Email::MIME> 1.861

=cut

sub new
{
    my( $this, $email, $config ) = @_;
    my $class = ref($this) || $this;

    my $c = bless
    {
     email => $email,
     plugin => {},
     config => $config,
    }, $class;

    unless( %PLUGIN )
    {
	&_register_plugins;
    }

    return $c;
}


#######################################################################

sub _register_plugins
{
#    warn "Classifiers are:";

    foreach my $plugin ( __PACKAGE__->plugins )
    {
	if( $plugin =~ /.*::(.*)/ )
	{
	    my $name = lc($1);

#	    warn " * ".$name;
	    $PLUGIN{$name} = $plugin;

	    foreach my $fname ( $name, 'is_'.$name )
	    {
		if( my $glob = $Email::Classifier::{$fname} )
		{
		    if( *{$glob}{'CODE'} )
		    {
			die "   Plugin $name conflicts with local function";
		    }
		}
	    }
	}
    }
}


#######################################################################

sub _import
{
    my( $c, $name ) = @_;

    if( my $plugin = $PLUGIN{$name} )
    {
#	warn "Classifier imports $name";

	no strict 'refs'; # Symbolic refs

	my $mod = package_to_module($plugin);
	require($mod);

	my $cp = $plugin->new($c->{'email'});
	$c->{'plugin'}{$name} = $cp;


	*{"Email::Classifier::$name"} = sub
	{
	    my( $c ) = @_;
#	    warn "Called $name for $c";
	    return $c->{'plugin'}{$name} ||=
	      $PLUGIN{$name}->new($c->{'email'},
				  $c->{'config'});
	};

	my $is_name = "is_$name";
	*{"Email::Classifier::$is_name"} = sub
	{
	    my( $c ) = shift;
#	    warn "Called $is_name for $c";
	    my $obj = $c->{'plugin'}{$name} ||=
	      $PLUGIN{$name}->new($c->{'email'},
				  $c->{'config'});
	    return $obj->$is_name(@_);
	};



#	warn "  Checking result of import";
#	if( my $glob = $Email::Classifier::{$name} )
#	{
#	    if( *{$glob}{'CODE'} )
#	    {
#		warn "   Plugin $name imported";
#	    }
#	}

	return 1;
    }

    return 0;
}


#######################################################################

AUTOLOAD
{
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;
    return if $method =~ /DESTROY$/;

    my( $c ) = shift;
    if( $method =~ /^is_(.*)/ )
    {
	my $name = $1;
	if( $c->_import($name) )
	{
	    return $c->$name->$method;
	}
    }
    else
    {
	if( $c->_import($method) )
	{
	    return $c->{'plugin'}{$method};
	}
    }

    croak sprintf qq(Can't locate object method "%s" via package "%s"), $method, __PACKAGE__;
}


#######################################################################

=head2 package_to_module

  package_to_module( $package )

Returns the correspondoing praxis filename of the module.

=cut

sub package_to_module
{
    my $package = shift;
    $package =~ s/::/\//g;
    $package .= ".pm";
    return $package;
}


#######################################################################

=head1 AUTHOR

Jonas Liljegren, C<< <jonas at cpan.org> >>

=head1 BUGS

Please report any bugs or feature requests to C<bug-email-classifier at rt.cpan.org>, or through
the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Email-Classifier>.  I will be notified, and then you'll
automatically be notified of progress on your bug as I make changes.




=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Email::Classifier


You can also look for information at:

=over 4

=item * RT: CPAN's request tracker

L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Email-Classifier>

=item * AnnoCPAN: Annotated CPAN documentation

L<http://annocpan.org/dist/Email-Classifier>

=item * CPAN Ratings

L<http://cpanratings.perl.org/d/Email-Classifier>

=item * Search CPAN

L<http://search.cpan.org/dist/Email-Classifier>

=back


=head1 ACKNOWLEDGEMENTS


=head1 COPYRIGHT & LICENSE

Copyright 2008 Jonas Liljegren, all rights reserved.

This program is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.


=cut

1; # End of Email::Classifier
