#  $Id: Bounce.pm,v 1.1.1.1 2008/02/17 23:31:10 jonas Exp $  -*-cperl-*-
package Email::Classifier::Bounce;
#=====================================================================
#
# AUTHOR
#   Jonas Liljegren   <jonas@cpan.org>
#
# COPYRIGHT
#   Copyright (C) 2008 Avisita AB.  All Rights Reserved.
#
#=====================================================================

=head1 NAME

Email::Classifier::Bounce

=cut

use strict;
use Carp qw( croak confess cluck );
use Data::Dump qw( dump );
use Module::Pluggable
  require => 1,
  search_path => ['Email::Classifier::Bounce'];

our $DEBUG = 0;


=head1 DESCRIPTION

Based on L<Mail::DeliveryStatus::BounceParser>

std_reason:

  user_unknown
  domain_error
  over_quota
  unknown

type:

  unknown
  hard
  soft
  transient
  block

=cut

our $EMAIL_ADDR_REGEX = qr{
# Avoid using something like Email::Valid
# Full rfc(2)822 compliance isn't exactly what we want, and this seems to work
# for most real world cases
(?:<|^|\s)            # Space, or the start of a string
([^\s\/<]+            # some non-space, non-/ characters; none are <
\@                    # at sign (duh)
(?:[-\w]+\.)+[-\w]+)  # word characters or hypens organized into
                      # at least two dot-separated words
(?:$|\s|>)            # then the end
}sx;


my $Returned_Message_Below = qr/(
    (?:original|returned) \s message \s (?:follows|below)
  | (?: this \s is \s a \s copy \s of
      | below \s this \s line \s is \s a \s copy
    ) .{0,100} \s message
  | message \s header \s follows
  | ^ (?:return-path|received|from):
)/sixm;

my $Common_Subjects = qr/ failure[ ]?notice
                        | undeliverable
                        | delivery[ ]?failure
                       /ix;

my @REPORT_FIELDS = qw(
 email
 host
 smtp_code
 reporting-mta
 arrival-date
 reason
 std_reason
);
my %REPORT_FIELDS; for(@REPORT_FIELDS){$REPORT_FIELDS{$_}=1};

#######################################################################

=head2 new

=cut

sub new
{
    my( $this, $email, $config ) = @_;
    my $class = ref($this) || $this;
    $config ||= {};

    my $c = bless
    {
     email => $email,
     reports => [],
     is_bounce => 0,
     guessed => 0,    # tried to find bounce info in plain text?
     type => undef,   # changed from Mail::DeliveryStatus::BounceParser
     effective_type => undef,
    }, $class;

    # backward compatibility...
    $c->{prefer_final_recipient} =
      $config->{prefer_final_recipient} || 0;


    # type: delivery-status $action
    #     : delivery-status expanded

    # Parsed on demand:
    #   orig_message
    #   orig_head
    #   orig_message_id

    $c->{'effective_type'} = lc($email->content_type || "text/plain");
    $c->{'effective_type'} =~ s/;.*//;

    $c->analyze;

    warn "Result: $c->{is_bounce}\n" if $DEBUG;

    if( $DEBUG )
    {
	for( my $i=0; $i<=$#{$c->{reports}}; $i++ )
	{
	    warn "  report $i:\n";
	    warn $c->{reports}[$i]->as_string;
	    warn "\n\n";
	}
    }

    return $c;
}


#######################################################################

=head2 is_bounce

=cut

sub is_bounce
{
    my( $c ) = @_;

    return $c->{'is_bounce'} ? 1 : 0;
}


#######################################################################

=head2 is_bounce_maby

=cut

sub is_bounce_maby
{
    my( $c ) = @_;

    $c->analyze_guess;

    foreach my $report ( @{$c->{reports}} )
    {
	my $std_reason = $report->header('std_reason') or next;
	return 1 if $std_reason ne 'unknown';
    }

    return 0;
}


#######################################################################

=head2 analyze

=cut

sub analyze
{
    my( $c ) = @_;

    my $top = $c->top;

    warn "  Analyzing for Bounce\n" if $DEBUG;

#    ### Check format
#    $c->p_ims;              # Internet Mail Service
#    $c->p_novell_groupwise; # Novell Groupwise
#    $c->p_plain_smtp_transcript;
#    $c->p_xdelivery_status;


    # $c->analyze_transient # rework this




    #
    # try to extract email addresses to identify members.
    # we will also try to extract reasons as much as we can.
    #

    warn "Message has type ".$c->effective_type."\n" if $DEBUG;


    if( $c->effective_type eq "multipart/report" )
    {
	warn "  analyzing multipart/report\n" if $DEBUG;

	my($delivery_status) =
	  $c->first_part_with_type("message/delivery-status");

	my %global =
	  (
	   "reporting-mta" => undef,
	   "arrival-date"  => undef,
	  );

	my( $seen_action_expanded, $seen_action_failed );


	# Some MTAs generate malformed multipart/report messages with
	# no message/delivery-status part
	my $delivery_status_body = $delivery_status->body || '';
	unless( $delivery_status_body )
	{
	    warn "  no message/delivery-status found\n" if $DEBUG;
	}


	# Used to be \n\n, but now we allow any number of newlines
	# between individual per-recipient fields to deal with stupid
	# bug with the IIS SMTP service.  RFC1894 (2.1, 2.3) is not
	# 100% clear about whether more than one line is allowed - it
	# just says "preceded by a blank line".  We very well may put
	# an upper bound on this in the future.
	#
	# See t/iis-multiple-bounce.t

	foreach my $para ( split /\n{2,}/, $delivery_status_body )
	{
	    # See t/surfcontrol-extra-newline.t - deal with bug #21249
	    $para =~ s/\A\n+//g;


#	    warn "REPORT:\n";
#	    warn $para."\n-------------\n";

	    my $report_in = $Email::Classifier::CLASS_HEADER->
	      new($para."\n");
	    my %report; # New report

	    if( $DEBUG )
	    {
		warn "Parsed a report:\n";
#		warn $report_in->as_string;
#		warn "-----------\n";
		warn join "", map{ "  $_\n"} split /\n/,
		  $report_in->as_string;
	    }


	    # Some MTAs send unsought delivery-status notifications
	    # indicating success; others send RFC1892/RFC3464 delivery
	    # status notifications for transient failures.

	    if( my $action = lc $report_in->header('Action'))
	    {
		if( $action =~ s/^\s*([a-z]+)\b.*/$1/s )
		{

		    # In general, assume that anything other than
		    # 'failed' is a non-bounce; but 'expanded' is
		    # handled after the end of this foreach loop,
		    # because it might be followed by another
		    # per-recipient group that says 'failed'.

		    if( $action eq 'expanded' )
		    {
			$seen_action_expanded = 1;
		    }
		    elsif( $action eq 'failed' )
		    {
			$seen_action_failed   = 1;
		    }
		    else
		    {
			warn("message/delivery-status says 'Action: $1'")
			  if $DEBUG;
			$c->{is_bounce} = 0;
			return 1;
		    }
		}
		else
		{
		    warn "Failed to parse action: $action\n" if $DEBUG;
		}
	    }
	    else
	    {
		warn "Report has no action\n" if $DEBUG;
	    }

	    foreach my $hdr (qw(reporting-mta arrival-date))
	    {
		if( my $report_hdr = $report_in->header($hdr) )
		{
		    warn "Extracting $hdr from report\n" if $DEBUG;
		    $global{$hdr} ||= $report_hdr;
		}

		if( $global{$hdr} )
		{
		    warn "Setting report $hdr = $global{$hdr}\n" if $DEBUG;
		    $report{$hdr} = $global{$hdr};
		}
	    }

	    my $rcpt; # recipient

	    my $rcpt_orig = $report_in->header("original-recipient");
	    my $rcpt_final = $report_in->header("final-recipient");

	    if( $c->{prefer_final_recipient} )
	    {
		$rcpt = $rcpt_final || $rcpt_orig || '';
	    }
	    else
	    {
		$rcpt = $rcpt_orig || $rcpt_final || '';
	    }

	    next unless $rcpt;

	    warn "Parsing for recipient: $rcpt\n" if $DEBUG;

	    # Diagnostic-Code: smtp; 550 5.1.1 User unknown
	    my $diag = $report_in->header("diagnostic-code") || '';

	    if( my $reason = $diag )
	    {
		$reason =~ s/[^;]+;\s*//; # strip leading X-Postfix;
		$report{reason} = $reason;
	    }

	    # strip leading RFC822; or LOCAL; or system;
	    $rcpt =~ s/[^;]+;\s*//;
	    $rcpt = $c->_cleanup_email($rcpt);
	    $report{email} = $rcpt;

	    if( my $status = $report_in->header('Status') )
	    {

		# RFC 1893... prefer Status: if it exists and is
		# something we know about.
		# Not 100% sure about 5.1.0...

		if( $status =~ /^5\.1\.[01]$/ )
		{
		    $report{std_reason} = "user_unknown";
		}
		elsif( $status eq "5.1.2" )
		{
		    $report{std_reason} = "domain_error";
		}
		elsif( $status eq "5.2.2" )
		{
		    $report{std_reason} = "over_quota";
		}
		else
		{
		    warn "Unknown status code: $status\n" if $DEBUG;
		    my $std_reason = _std_reason($diag);
		    $report{std_reason} = $std_reason;
		}
	    }
	    else
	    {
		warn "No status given in report\n" if $DEBUG;
		my $std_reason = _std_reason($diag);
		$report{std_reason} = $std_reason;
	    }

	    my($host) = $diag =~ /\bhost\s+(\S+)/i;
	    $report{host} = $host if $host;

	    warn "diag: $diag\n" if $DEBUG;
	    if( $diag =~ m/ ( [245] \d{2} ) \s
                          | \s ( [245] \d{2} ) (?!\.) /x )
	    {
		my $code = $1 || $2 || '';
		warn "Got code '$code'\n" if $DEBUG;
		$report{smtp_code} = $code if $code;
	    }


	    unless( $host )
	    {
		warn "Setting host based on recipient\n" if $DEBUG;
		$host = ($rcpt =~ /\@(.+)/)[0];
		$report{host} = $host if $host;
	    }

	    if( $report{smtp_code} =~ /^2../ )
	    {
		warn "SMTP code is $report{smtp_code}; no_problemo\n" if $DEBUG;
		next;
	    }

	    warn "Reason found in report: $report{std_reason}\n" if $DEBUG;

	    $c->new_report(\%report);

	} # END foreach $para


	if( $seen_action_expanded && !$seen_action_failed )
	{
	    # We've seen at least one 'Action: expanded' DSN-field,
	    # but no 'Action: failed'

	    warn("message/delivery-status says 'Action: expanded'\n")
	      if $DEBUG;
	    $c->{is_bounce} = 0;
	    return 1;
	}

	$c->{is_bounce} = 1;
    }
    else
    {
	$c->_do_plugins('analyze');
    }

    return 1 if $c->{is_bounce};


    # Only try to guess bounce report if we belive this realy could be
    # a bounce message

    if( $top->header('Auto-Submitted') or
	$top->header('From') =~ /mailer-daemon|postmaster/i or
	$top->header('Subject') =~ $Common_Subjects
      )
    {
	if( $c->is_bounce_maby )
	{
	    $c->{is_bounce} = 1;
	}
    }

    return 1;
}


#######################################################################

=head2 analyze_guess

Only if we think that it may be a bounce, but not in a standard format

=cut

sub analyze_guess
{
    my( $c ) = @_;
    my $top = $c->top;

    return 1 if $c->is_bounce; # Already found a better answer
    return if $c->{guessed};

    warn "  Analyzing for Bounce - guessing\n" if $DEBUG;

    if( $c->effective_type =~ /multipart/i )
    {
	# but not a multipart/report.  look through each non-message/*
	# section.  See t/corpus/exchange.unknown.msg

	my @delivery_status_parts =
	  grep{ $_->content_type =~ m{text/plain}i } $top->parts;

	warn "Trying to extract reports from multipart message\n"
	  if $DEBUG;

	foreach my $status_part ( @delivery_status_parts )
	{
	    my $text = $status_part->body;
	    push @{$c->{reports}}, $c->_extract_reports($text);
	}
    }
    elsif( $c->effective_type =~ m{text/plain}i )
    {
	# handle plain-text responses

	# This used to just take *any* part, even if the only part
	# wasn't a text/plain part
	#
	# We may have to specifically allow some other types, but in
	# my testing, all the messages that get here and are actual
	# bounces are text/plain wby - 20060907

	# they usually say "returned message" somewhere, and we can
	# split on that, above and below.

	warn "Trying to find report in single text/plain body\n"
	  if $DEBUG;

	my $body_string = $top->body;

	if( $body_string =~ $Returned_Message_Below )
	{
	    warn "Matching Returned_Message_Below\n" if $DEBUG;
	    my ($stuff_before, $stuff_splitted, $stuff_after) =
	      split $Returned_Message_Below, $body_string, 3;

	    push @{$c->{reports}}, $c->_extract_reports($stuff_before);

	    # TODO: Set up $c->{'orig_message'}
	}
	elsif( $body_string =~ /(.+)\n\n(.+?Message-ID:.+)/is )
	{
	    warn "Matching Message-ID string\n" if $DEBUG;
	    push @{$c->{reports}}, $c->_extract_reports($1);
	}
	else
	{
	    warn "  looking at the thole part\n" if $DEBUG;
	    push @{$c->{reports}}, $c->_extract_reports($body_string);
	}
    }

    return $c->{guessed} = 1;
}


#######################################################################

=head2 analyze_transient

=cut

sub analyze_transient
{
    my( $c ) = @_;

    my $top = $c->top;

    my $Not_An_Error = qr/
    \b delayed \b
  | \b warning \b
  | transient.{0,20}\serror
  | Your \s message .{0,100} was \s delivered \s to \s the \s following \s recipient
/six;

    my $Really_An_Error = qr/this is a permanent error/i;


    # nonfatal errors usually say they're transient

    if( $c->effective_type eq "text/plain"
	and $top->body =~ /transient.*error/is )
    {
	warn("seems like a transient error\n") if $DEBUG;
	$c->{type} = 'transient';
	return 1;
    }


    # nonfatal errors usually say they're transient, but sometimes
    # they do it straight out and sometimes it's wrapped in a
    # multipart/report.
    #
    # Be careful not to examine a returned body for the transient-only
    # signature: $Not_An_Error can match the single words 'delayed'
    # and 'warning', which could quite reasonably occur in the body of
    # the returned message.  This also means it's worth additionally
    # checking for a regex that gives a very strong indication that
    # the error was permanent.


    my $part_for_maybe_transient;

    if( $c->effective_type eq "text/plain" )
    {
	$part_for_maybe_transient = $top;
    }
    elsif( $c->effective_type =~ /multipart/i
	   and $c->effective_type ne 'multipart/report' )
    {
	($part_for_maybe_transient) =
	  $c->first_part_with_type("text/plain");
    }

    if( $part_for_maybe_transient )
    {
	my $string = $part_for_maybe_transient->body;

	# Stop checking if no sign of transient error
	my $transient_pos = _match_position($string, $Not_An_Error);
	return unless defined $transient_pos;

	warn "  Sign of a transient error\n" if $DEBUG;

	my $permanent_pos = _match_position($string, $Really_An_Error);
	my $orig_msg_pos  = _match_position($string, $Returned_Message_Below);

	# Stop checking if PERMANENT ERROR in report
	if( _position_before($permanent_pos, $orig_msg_pos) )
	{
	    warn "    looks like a PERMANENT error\n" if $DEBUG;
	    $c->{'is_bounce'} = 1;
	    return;
	}

	if( _position_before($transient_pos, $orig_msg_pos) )
	{
	    warn("  transient error\n") if $DEBUG;
	    $c->{type} = 'transient';
	    return 1;
	}
    }

    return 0;
}


#######################################################################

=head2 top

=cut

sub top
{
    return $_[0]->{'email'};
}


#######################################################################

=head2 reports

=cut

sub reports
{
    return @{$_[0]->{'reports'}};
}


#######################################################################

=head2 new_report

=cut

sub new_report
{
    my( $c, $data ) = @_;
    $data ||= {};

    my $report = $Email::Classifier::CLASS_HEADER->new("");

    foreach my $key (keys %$data)
    {
	unless( $REPORT_FIELDS{$key} )
	{
	    dump(\%REPORT_FIELDS);
	    confess "Not a valid report field: $key";
	}

	my $val = $data->{$key};
	$report->header_set($key => $val);
	warn "  $key: $val\n" if $DEBUG;
    }

    push @{$c->{'reports'}}, $report;
    $c->{'is_bounce'} = 1;
    return $report;
}


#######################################################################

=head2 addresses

Returns a list of the addresses which appear to be bouncing.  Each
member of the list is an email address string of the form
'foo@bar.com'.

=cut

sub addresses
{
#    my @res;
#    foreach my $report ( shift->reports )
#    {
#	warn "report $report";
#	my @emls = $report->header("email");
#	warn "emails @emls";
#	push @res, @emls;
#    }

    my(@res) = map { $_->header("email") } shift->reports;
#    warn "Returning addresses @res";

    return(@res);
}


#######################################################################

=head2 effective_type

=cut

sub effective_type
{

    return $_[0]->{'effective_type'};
}


#######################################################################

=head2 first_part_with_type

=cut

sub first_part_with_type
{
    my( $c, $type ) = @_;

    foreach my $part ( $c->top->parts )
    {
	if( ($part->content_type||'text/plain') =~ /^$type/i )
	{
	    return $part;
	}
    }

    return undef;
}


#######################################################################

=head2 first_non_multi_part

=cut

sub first_non_multi_part
{
    my( $c ) = @_;

    my $part = $c->top;
    while( ($part->content_type||'text/plain') =~ /^multipart\//i )
    {
	( $part ) = $part->parts;
    }

    return $part;
}


#######################################################################

=head2 orig_message

=cut

sub orig_message
{
    my( $c ) = @_;

    if( exists $c->{orig_message} )
    {
	return $c->{orig_message};
    }

    warn "Trying to find an original message part\n" if $DEBUG;

    # "Internet Mail Service" sends multipart/mixed which still
    # has a message/rfc822 in it

    if( my $part = $c->first_part_with_type("message/rfc822") )
    {
	warn "Extracted original message\n" if $DEBUG;
	my $env = $c->_envelope( $part );

	$c->{orig_head} = $env->header_obj;
	$c->{orig_message_id} = $env->header("message-id");
	unless( $c->{orig_message_id} )
	{
	    warn "Failed to get message-id from rfc822 header:\n";
	    warn $env->header_obj->as_string;
	    warn "\n";
	}

	return $c->{orig_message} = $env;
    }

    return $c->{orig_message} = undef;
}


#######################################################################

=head2 orig_head

=cut

sub orig_head
{
    my( $c ) = @_;

    if( exists $c->{orig_head} )
    {
	return $c->{orig_head};
    }

    if( $c->orig_message ) # sets head
    {
	return $c->{orig_head};
    }

    warn "Trying to find an original message head\n" if $DEBUG;

    # "Internet Mail Service" sends multipart/mixed which still
    # has a message/rfc822 in it

    if( my $part = $c->first_part_with_type("text/rfc822-headers") )
    {
	my $data_dec = $part->body;
	my $msg;

	my $head = $Email::Classifier::CLASS_HEADER->new(\$data_dec );
	if( $head )
	{
	    warn "Extracted original message head\n" if $DEBUG;
	    $c->{orig_message_id} = $head->header("message-id");
	    return $c->{orig_head} = $head;
	}
	else
	{
	    warn "Failed to parse part head\n" if $DEBUG;
	}
    }

    return $c->{orig_head} = undef;
}


#######################################################################

=head2 orig_message_id

=cut

sub orig_message_id
{
    my( $c ) = @_;

    if( exists $c->{orig_message_id} )
    {
	return $c->{orig_message_id};
    }

    if( $c->orig_head ) # sets message-id
    {
	return $c->{orig_message_id};
    }

    # Tries to find message-id in body
    my $part = $c->first_non_multi_part;
    if( $part->body =~ /Message-ID: (\S+)/i )
    {
	$c->{orig_message_id} = $1;
	warn "found a message-id $c->{orig_message_id} in the body\n"
	  if $DEBUG;
	return $c->{orig_message_id};
    }

    return $c->{orig_message_id} = undef;
}


#######################################################################

# Return the position in $string at which $regex first matches, or
# undef if no match.

sub _match_position
{
    my( $string, $regex ) = @_;
    return $string =~ $regex ? $-[0] : undef;
}


#######################################################################

sub _position_before
{
    my( $pos_a, $pos_b ) = @_;
    return 1 if defined($pos_a) && (!defined($pos_b) || $pos_a < $pos_b);
    return 0;
}


#######################################################################

sub _cleanup_email
{
    my $c = shift;
    local $_ = shift;

    chomp;
    # Get rid of parens around addresses like (luser@example.com)
    # Got rid of earlier /\(.*\)/ - not sure what that was about - wby
    tr/[()]//d;
    s/^To:\s*//i;
    s/[.:;]+$//;
    s/<(.+)>/$1/;
    # IMS hack: c=US;a= ;p=NDC;o=ORANGE;dda:SMTP=slpark@msx.ndc.mc.uci.edu; on
    # Thu, 19 Sep...
    s/.*:SMTP=//;
    s/^\s+//;
    s/\s+$//;
    # hack to get rid of stuff like "luser@example.com...User"
    s/\.{3}\S+//;
    # SMTP:foo@example.com
    s/^SMTP://;

    return $_;
}


#######################################################################

sub _std_reason
{
    local $_ = shift;

    if( /(?:domain|host)\s+(?:not\s+found|unknown)/i )
    {
	return "domain_error";
    }

    if(
       /try.again.later/is or
       /mailbox\b.*\bfull/ or
       /storage/i          or
       /quota/i            or
       /\s552\s/           or
       /\s#?5\.2\.2\s/                                  # rfc 1893
      )
    {
	return "over_quota";
    }

    my $user_re =
      qr'(?: mailbox  | user | recipient | address (?: ee)?
          | customer | account | e-?mail | <? $EMAIL_ADDR_REGEX >? )'ix;

    if(
       /\s \(? \#? 5\.1\.[01] \)? \s/x or                  # rfc 1893
       /$user_re\s+ (?:\S+\s+)? (?:is\s+)?                 # Generic
         (?: (?: un|not\s+) (?: known | recognized )
           | [dw]oes\s?n[o']?t
           (?: exist|found ) | disabled | expired ) /ix or
       /no\s+(?:such\s+)?$user_re/i or                     # Gmail and other
       /inactive user/i or                                 # Outblaze
       /unknown local part/i or                            # Exim(?)
       /user\s+doesn't\s+have\s+a/i or                     # Yahoo!
       /account\s+has\s+been\s+(?:disabled|suspended)/i or # Yahoo!
       /$user_re\s+(?:suspended|discontinued)/i or         # everyone.net / other?
       /unknown\s+$user_re/i or                            # Generic
       /$user_re\s+(?:is\s+)?(?:inactive|unavailable)/i or # Hotmail, others?
       /(?:(?:in|not\s+a\s+)?valid|no such)\s$user_re/i or # Various
       /$user_re\s+(?:was\s+)?not\s+found/i or             # AOL, generic
       /$user_re \s+ (?:is\s+)? (?:currently\s+)?
         (?:suspended|unavailable)/ix or                   # ATT, generic
       /address is administratively disabled/i or          # Unknown
       /no $user_re\s+(?:here\s+)?by that name/i or        # Unknown
       /<?$EMAIL_ADDR_REGEX>? is invalid/i or              # Unknown
       /address.*not known here/i or                       # Unknown
       /recipient\s+(?:address\s+)?rejected/i or           # Cox, generic
       /not\s+listed\s+in\s+Domino/i or                    # Domino
       /account not activated/i or                         # usa.net
       /not\s+our\s+customer/i or                          # Comcast
       /doesn't handle mail for that user/i                # mailfoundry
      )
    {
	return "user_unknown";
    }

    if(
       /domain\s+syntax/i or
       /timed\s+out/i or
       /route\s+to\s+host/i or
       /connection\s+refused/i or
       /no\s+data\s+record\s+of\s+requested\s+type/i
      )
    {
	return "domain_error";
    }

    return "unknown";
}


#######################################################################

sub _extract_reports
{
    my( $c, $text ) = @_;

    my %by_email;

    # we'll assume that the text is made up of:
    # blah blah 0
    #             email@address 1
    # blah blah 1
    #             email@address 2
    # blah blah 2
    #

    # we'll break it up accordingly, and first try to detect a reason
    # for email 1 in section 1; if there's no reason returned, we'll
    # look in section 0.  and we'll keep going that way for each
    # address.

    return unless $text;

    my @split = split($EMAIL_ADDR_REGEX, $text);

    foreach my $i ( 0 .. $#split )
    {
	# only interested in the odd numbered elements, which are the
	# email addressess.
	next if $i % 2 == 0;

	my $email = $c->_cleanup_email($split[$i]);

#	if( $split[$i-1] =~ /they are not accepting mail from/ )
#	{
#	    # aol airmail sender block
#	    warn "aol airmail sender block\n" if $DEBUG;
#	    next;
#	}

	my $std_reason = "unknown";
	$std_reason = _std_reason($split[$i+1]) if $#split > $i;
	$std_reason = _std_reason($split[$i-1]) if $std_reason eq "unknown";

	# todo:
	# if we can't figure out the reason, if we're in the
	# delivery-status part, go back up into the text part and try
	# extract_report() on that.


	# Keep the best reason found so far:
	if( exists $by_email{$email}
	    and
	    $by_email{$email}->{std_reason} ne "unknown"
	    and
	    $std_reason eq "unknown"
	  )
	{
	    next;
	}

	$by_email{$email} =
	{
	 email => $email || '',
	 raw   => join ("", @split[$i-1..$i+1]),
	 std_reason => $std_reason || '',
	};
    }


#    warn "reports to generate:";
#    warn datadump(\%by_email);

    my @toreturn;
    foreach my $email (keys %by_email)
    {
	my $report = $Email::Classifier::CLASS_HEADER->new("");
#	warn "Creating a new report";
	foreach my $key ( keys %{$by_email{$email}} )
	{
#	    warn "  setting $key = $by_email{$email}{$key}";
	    $report->header_set($key => $by_email{$email}{$key});
	}

	push @toreturn, $report;
    }

    return @toreturn;
}


#######################################################################

sub _do_plugins
{
    my $c = shift;
    my $method = shift;

    warn "Running $method plugins\n" if $DEBUG;
    no strict 'refs'; # Symbolic refs
    my $res;
    foreach my $plugin ( $c->plugins )
    {
	if( $plugin->can($method) )
	{
	    warn "   running $plugin\n" if $DEBUG;
	    last if $res = &{$plugin."::".$method}($c, @_);
	}
    }
    return $res;
}


#######################################################################

sub _envelope
{
    my( $c, $part ) = @_;

    if( $part->can('envelope') )
    {
	return $part->envelope;
    }
    else
    {
	my $class = ref $part;
	return $class->new( $part->body );
    }
}


#######################################################################

1;


__END__

Auto Reply (AR): An automatic response from the recipient.
Examples: "Out of Office", "Vacation Message"

Bounce With No Email Address (BN): eList Manager determined that the message was a bounce, but it could not resolve the original email recipient email address. Rare.

Challenge-Response Message (CR): Challenge-Response email systems were created as a reaction to the increasing circulation of spam. Dubbed as a 100% effective solution to stopping spam from ever reaching users' inboxes, Challenge-Response systems like EarthLink SpamBlocker and SpamArrest, among others, require human intervention for an email message to reach its intended recipient.

In short, a Challenge-Response message is an automatic response from the recipient, requesting that the sender confirm a real person is sending the message. Generally, confirmation is completed manually by clicking on a hyperlink within the Challenge-Response message itself.

When eList Manager determines a message is a Challenge-Response type message, the email address that is returned is the FROM: email address of the system, or person, sending the Challenge-Response message to you. It's been our experience that the original email address of the recipient does not appear within the Challenge-Response message, thus there is no way to retrieve the original email address. You'll be handling these types of messages manually, because that's what they're designed for, so it should be okay that the original email address isn't returned.

General Bounce (GB): The email server could not deliver your email message, but the bounce message was unclear as to what kind of bounce it was. Most eList Manager users treat these as soft bounces.

Example: "Subject: Undeliverable mail"

Hard Bounce (HB): The email server could not deliver your email message.

Example: "550 User Unknown"

Mail Block - General (MB): Indicates that the recipient's email server is blocking email from your email server.

Example: "550 Message REFUSED by peer"

Mail Block - Known Spammer (MBKS): Indicates that the recipient's email server is blocking email from your email server because it believes you are a known spammer.

Example: "REJECT Known SPAM source"

Mail Block - Spam Detected (MBSD): Indicates that the recipient's email server is blocking email from your email server because the message appears to have content that looks like spam.

Example: "550 Possible spam detected"

Mail Block - Attachment Detected (MBAD): Indicates that the recipient's email server is blocking email from your email server because the email message contained an attachment.

Example: "552 Disapproved attachment"

Mail Block - Relay Denied (MBRD): Indicates that the recipient's email server is blocking email from your email server.

Example: "551 relaying denied"

Non Bounce (NB): eList Manager determined that the message was not a bounce. This could be a recipient reply, or maybe a bounce format that eList Manager did not recognize.

Soft Bounce - General (SB): The email server is temporarily unable to deliver your message to the recipient email address.

Example: "Connection timed out."

Soft Bounce - Dns Failure (SBDF): The email server is temporarily unable to deliver your message to the recipient email address because of a DNS problem.

Example: "Host is unreachable"

Soft Bounce - Mailbox Full (SBMF): The email server is temporarily unable to deliver your message to the recipient email address because the recipient's email box is full.

Example: "Mailbox over quota"

Soft Bounce - Message Size Too Large (SBMS): The email server could not deliver your message to the recipient because the message size is too large.

Example: "Exceeded maximum inbound message size"

Subscribe Request (SR): Someone is requesting to be added to your opt-in email list. eList Manager looks for subscribe requests in the subject header of the email message, and on the first line of a text based email message. eList Manager parses for subscribe requests in several different 7-bit ascii langauges. (ie. English, Spanish, French, German, etc.)

Transient Bounce (TB): The email server temporarily can not deliver your message, but it is still trying.

Examples: "Warning: message still undelivered after 4 hours. Will keep trying until message is 2 days old"

Unsubscribe Request (UR): Recipient is requesting to be removed from future email from you. eList Manager looks for unsubscribe/remove requests in the subject header of the email message, and on the first line of a text based email message. eList Manager parses for unsubscribe requests in several different 7-bit ascii langauges. (ie. English, Spanish, French, German, etc.)

.....

Recognize virus email notifications.

Recognize email address change notification messages.

